
# -----------------------------------------------
#  EMMIX-WIRE, Version 1.0.1;  (October 12,2009)
#  updated by Angus (15/1/2018): change EMMIXskew to emmixskew
#  updated by Angus (15/1/2018): add output of estimated profiles
# -----------------------------------------------

# new features:

# 1. incoporate the time dependecy with AR(1) type of random gene effects.
# 2. for time series gene profiles, periods can be different for each clusters

#----------------------------------------------------------

require(emmixskew)

# do the inverse of R
ar.get.A<-function(phi,theta,p,g)  #
{

	#calcaulate J, K

	Ap<-diag(0,p)

      if(p>1)
	for(i in 1:(p-1))
		Ap[i,i+1]<-1
	J<-t(Ap)+Ap

	Bp<-diag(0,p)

	if(p>1)
	{
		Bp[1 , 1]<-1
		Bp[p,p]<-1
	}
	else
	Bp[1,1]<-2

	K<-Bp

DU <- IA <- array(0,c(p,p,g))

for( h in 1:g){

IR  <- as.matrix( (1+phi[h]^2)*diag(p)-phi[h]*J-phi[h]^2*K )

DU[,,h] <- theta[h] * solve(IR)

IA[,,h] <- IR/theta[h]
}
#

list(DU=DU,IA=IA)
}


ar.L1L2L3<-function(TU,ni)
{

	L1    <- sum(diag(TU))

	L2<-L3<- 0

	if(ni > 1)
	L2<-(L2+sum(diag((rbind(TU[2:ni,],0)+rbind(0,TU[1:(ni-1),])))))

	L3<-L3+TU[1,1]+TU[ni,ni]

	L2<-L2/2


#  L1 = tr(x I), L2 = tr(x J), L3 = tr(x K)


c(L1,L2,L3)
}


#get k th (up to 4) order Fourier Series extension

get.X <- function(tim,period,k,one=FALSE)
{

X <- cbind(cos(2*pi*tim/period), sin(2*pi*tim/period))

if(k > 1 ) {
for(j in 2: (min(4,k)) )
X <- cbind(X,cos(2*j * pi*tim/period), sin(2*j* pi*tim/period))
}

if(one)
cbind(1,X)
else
X
}


#a function to get the patition from the posterir probability
tau2clust<-function(tau)
{
apply(tau,FUN=which.max,MARGIN=1)
}

#a function to normalize the rows
DoRows<-function(dat)
{
dat<-as.matrix(dat)
(dat-rowMeans(dat,na.rm=T))/sqrt(apply(dat,FUN=var,1))
}

#a function to normalize the columns
DoCols<-function(dat)
{
t(DoRows(t(dat)))
}


#----------------------------------

clust.plot.time<-function(clust,dat,base=10,nr=4)
{

# part 1: read the data

n<-nrow(dat);
m<-ncol(dat)
K<-max(clust)

time<-seq(from=0,by=10,length=m)

# part 2: plot
#X11()
par(mfrow=c(nr,2),mar=c(2,4,1,1)+0.2)
for(h in 1:K)
{

	if(sum(clust==h)==0)
		{
		plot(time,seq(from=min(dat)-0.1,to=max(dat)+0.1,length=m),type="n",xlab=""
			,ylab=paste("cluster ",h,"(",sum(clust==h),"genes )")               )
		}
	else
 	{
		if( sum(clust==h)>1)
			mat<-t( (dat)[clust==h,])
		else
			mat<-(dat)[clust==h,]
	#print(mat)
	plot(time,seq(from=-base-0.1,to=base+0.1,length=m),type="n",xlab=""
		,ylab=paste("cluster ",h,"(",sum(clust==h),"genes )")               )
	matlines(time,mat,col=h,lty=1)

	#lines(time,mean[,h],type=)
	if(h!=K&h%%(nr*2)==0)
	{
#	X11()  #windows(record=T)
	par(mfrow=c(nr,2),mar=c(2,4,1,1)+0.2)
	}
	}


}

#return(0)
}



#----------------------------

ar.init.fit<-function(dat,xsxc,qe,n,m,g,nkmeans)
{

found<-NULL
found$loglik<--Inf


for(j in 1:nkmeans)
{

initobj<-try(ar.init.km(dat,xsxc,qe,n,m,g))

if(length(initobj)==4)
if(initobj$loglik>found$loglik)
found<-initobj

} #end of loop

found
}

ar.init.km<-function(dat,xsxc,qe,n,m,g)
{
#
cluster<-rep(1,n)

if(g>1)
cluster<- kmeans(dat,g,nstart=5)$cluster
ar.init.reg(dat,xsxc,qe,n,m,g,cluster)
}

ar.init.reg<-function(dat,xsxc,qe,n,m,g,clust)
{
# set x for regression
x<-as.matrix(xsxc)

beta<-matrix(0,nrow=ncol(x),ncol=g)

sigma<-pro<-rep(0,g)

for( ij in 1:g)
{
ni<-sum(clust==ij)

nn     <- min(ni,10000%/%m)
select <- sample(1:ni,nn)


y      <- c(t(dat[clust==ij,][select,]))

if(nn > 1)
{
for(i in 2:nn)
x <- rbind(x,xsxc)
}

obj<-lm(y~0+x)
beta[,ij]<-coef(obj)
sigma[ij]<-sum((resid(obj))^2)/nn/m

pro[ij]<-ni/n

x<-as.matrix(xsxc) # reset x for next iteration

}



# loglikelihood

mu     <- array(0,c(m,g))
msigma <- array(0,c(m,m,g))

for(h in 1:g) {

mu[,h] <-c(xsxc%*%beta[,h])

msigma[,,h] <- diag(sigma[h],m)

}

obj <- tau.estep.wire(dat,pro,mu,msigma,n,m,g)

loglik <- obj$loglik

thete<-matrix(0,ncol=g,nrow=qe)
for(i in 1:qe)
thete[i,]<-sigma

list(beta=beta,thete=thete,pro=pro,loglik=loglik,mu=mu)

}

#esteps and msteps

mat.ABC.wire<-function(U,VV,W,thete,DU,thetv,ng,m)
{
A<-B<-C<-BC<-array(0,dim=c(m,m,ng))

for(h in 1:ng)
{

if(ncol(W)>1)
A[,,h]<-diag(as.vector(W%*%thete[,h]))
else
{
A[,,h]<-diag(c(W[,1]*thete[1,h]))

}
B[,,h]<-A[,,h]+U%*%DU[,,h]%*%t(U)
C[,,h]<-thetv[h]*VV
BC[,,h]<-B[,,h]+C[,,h]
}
list(A=A,B=B,C=C,BC=BC)
}

#
matM.wire<-function(B,C,tau,ng,m)
{
M<-array(0,dim=c(m,m,ng))
for(h in 1:ng)
M[,,h]<-solve(B[,,h]+C[,,h]*sum(tau[,h]))
M
}


#E-step
tau.estep.wire<-function(dat,pro,mu,sigma,n,p,g)
{

obj <- .Fortran("estepmvn",PACKAGE="emmixskew",
as.double(dat),as.integer(n),as.integer(p),as.integer(g),
as.double(pro),as.double(mu),as.double(sigma),
mtauk=double(n*g),double(g),loglik=double(1),
error = integer(1))[8:11]

if(obj$error) stop("error")

tau <- array(obj$mtauk,c(n,g))

list(tau=tau,loglik=obj$loglik,pro=colMeans(tau))

}



#--------------------------------------------

eb.estep<-function(tau,y,mu,DU,U,B,C,M,ng,n,m,qb)
{
eb1<-array(0,dim=c(n,qb,ng))
eb2<-array(0,c(ng,qb,qb))
invB<-array(0,c(m,m,ng))

for(h in 1:ng)
{
invB[,,h]<-solve(B[,,h])
# E(bi|y)

eb1[,,h]<-t( DU[,,h]%*%t(U)%*%invB[,,h]%*%(  (t(y)-mu[,h])
-c(C[,,h]%*%M[,,h]%*%colSums(t(t(y)-mu[,h])*tau[,h])) ))
#E(bi%*%bi^t|y)
eb2[h,,]<-( sum(tau[,h])*DU[,,h]-(sum(tau[,h])-1)*DU[,,h]%*%t(U)%*%invB[,,h]%*%U%*%DU[,,h]
-DU[,,h]%*%t(U)%*%M[,,h]%*%U%*%DU[,,h])
DU[,,h]<-(eb2[h,,]+ t(eb1[,,h]*tau[,h])%*%eb1[,,h])/sum(tau[,h])
}

list(DU=DU,eb1=eb1,invB=invB)
}

#-------------------------------


ar.wire.estep<-function(tau,y,mu,DU,B,C,M,n,m,g)
{
eb1 <-array(0,dim=c(n,m,g))
eb2 <-array(0,c(g,m,m))
invB<- TU <- array(0,c(m,m,g))

U <- diag(m)

for(h in 1:g)
{
invB[,,h]<-solve(B[,,h])
# E(bi|y)

eb1[,,h]<-t( DU[,,h]%*%t(U)%*%invB[,,h]%*%(  (t(y)-mu[,h])
-c(C[,,h]%*%M[,,h]%*%colSums(t(t(y)-mu[,h])*tau[,h])) ))
#E(bi%*%bi^t|y)
eb2[h,,]<-( sum(tau[,h])*DU[,,h]-(sum(tau[,h])-1)*DU[,,h]%*%t(U)%*%invB[,,h]%*%U%*%DU[,,h]
-DU[,,h]%*%t(U)%*%M[,,h]%*%U%*%DU[,,h])

TU[,,h]<-(eb2[h,,]+ t(eb1[,,h]*tau[,h])%*%eb1[,,h])

}

list(TU=TU,eb1=eb1,invB=invB)
}

ar.wire.mstep <- function(p,g,TU,sumtau) {

phi <- theta <- rep(0,g)


for(h in 1:g)
{

ret <- ar.L1L2L3(TU[,,h],p)
L1 <- ret[1]
L2 <- ret[2]
L3 <- ret[3]

#
c1 <- (p-1)*(L1-L3)
c2 <- (2-p)*L2
c3 <- p*L3-(1+p)*L1
c4 <- p*L2
#

cubif<-function(x,c1,c2,c3,c4){c1*x*x*x+c2*x*x+c3*x+c4}

px  <-(uniroot(cubif,c(-1,1),c1=c1,c2=c2,c3=c3,c4=c4))$root

th  <- 0
if(sumtau[h]>3)
th  <- ( L1-2*px*L2+ px^2*(L1-L3) )/p/sumtau[h]

phi[h]   <- px
theta[h] <- th
}

list(phi=phi,theta=theta)

}

#---------------------------

ec.estep<-function(tau,y,mu,thetv,V,M,ng,n,qc)
{
ec1<-array(0,c(qc,ng))
ec2<-kc<-rep(0,ng)
#
for(h in 1:ng)
{
ec1[,h]<-thetv[h]*t(V)%*%M[,,h]%*%colSums(t(t(y)-mu[,h])*tau[,h])
kc[h] <-thetv[h]*qc-sum(diag(t(V)%*%M[,,h]%*%V))*(thetv[h]^2*sum(tau[,h]))
ec2[h]<-c(t(ec1[,h])%*%ec1[,h]+kc[h])/qc
}
#
list(ec2=ec2,ec1=ec1)
}

ee.estep<-function(y,mu,tau,U,V,W,A,invB,M,ng,n,m,qe,dw,eb1,ec1)
{
#
ae<-ee<-array(0,dim=c(n,m,ng))
ke<-rep(0,ng)
thet<-matrix(0,ncol=ng,nrow=qe)
mi<-diag(t(W)%*%W)

#
for(h in 1:ng)
{


#    alignment

ae[,,h]<-t(mu[,h]+U%*%t(eb1[,,h])+c(V%*%ec1[,h]))

ee[,,h]<- (y-ae[,,h])
#
#ee[,,h]<-t( t(y)-mu[,h]-U%*%t(eb1[,,h])-c(V%*%ec1[,h]))


for(id in 1:ncol(W))
{
AL<-A[,,h]*W[,id]
ke[h]<-(sum(tau[,h])*sum(diag(AL))-sum(diag(  t(AL)%*%M[,,h]%*%AL))
-(sum(tau[,h])-1)*sum(diag( t(AL)%*%invB[,,h]%*%AL)))

thet[id,h]<-( sum(c(  t((t(ee[,,h])*W[,id])^2)*tau[,h]  ))+ke[h])/(mi[id]*sum(tau[,h]))
}# end of loop

} #end of loop
list(thete=thet,ae=ae)
}

get.period <- function(y,beta,g,A,tau,tmpoints,order,one,periods,U,V,eb1,ec1)
{

#periods is a vector of possible periods to choose from.

bestperiod <- rep(NA,g)

for(h in 1:g)
{
lok <- -Inf

# U b + V c
ae  <-  (U%*%t(eb1[,,h])+c(V%*%ec1[,h]))

inv <-  1/c(diag(A[,,h]))

ww  <-  tau[,h]
#

for(j in 1:length(periods)) {

X      <- get.X(tmpoints,periods[j],order,one)
mu     <- as.vector(X%*%beta[,h])
#
ee  <-  y - t(ae + mu)

ss  <-  - sum(diag( t(ee*ww)%*%ee  )*inv)

cat("\nss=",h, j, ss,periods[j], bestperiod[h])

if(ss > lok) {
lok <- ss
bestperiod[h] <- periods[j]
}

} #end loop pd

} #end loop h


bestperiod
} #end function


#---------------------------------

#  EMMIX-WIRE

#---------------------------------


# main functions


# start EMMIX-WIRE analysis from initial values
fit.emmix.wire<-function(dat,X,W,U,V,
pro,beta,thete,thetu,thetv,theta,phi,
n,m,g,nb,qb,qc,qe,
debug,ncov,nvcov,itmax,epsilon,
tmtype,tim,periods,k,one,bestperiod,common,mu,eb1,ec1)
{


# controls
flag<-error<-0


lk<-rep(0,itmax)
xxx <- array(0,c(nb,m,g))

VV<-V%*%t(V)
dw<-diag(t(W)%*%W)

if(tmtype){
periods <- c(periods)

} else {
for(h in 1:g)
xxx[,,h]<-solve(t(X)%*%X)%*%t(X)
}

mu<-matrix(0,ncol=g,nrow=m)

#-----------------------------------

for(i in 1:itmax)
{

if(ncov =="AR")
thetu   <- (ar.get.A(phi,theta,m,g))$DU

mobj<-mat.ABC.wire(U,VV,W,thete,thetu,thetv,g,m)
A<-mobj$A
B<-mobj$B
C<-mobj$C
BC<-mobj$BC

#-------------------------


if(tmtype) {

# if tmtype==TRUE,  update the mu and xxx

for(h in 1:g) {

X          <- get.X(tim,bestperiod[h],k,one)

xxx[,,h]   <- solve(t(X)%*%X)%*%t(X)

mu[,h]     <- as.vector(X%*%beta[,h])

}  #end loop h

#----------------------------

} else {

for(h in 1:g) {

mu[,h]     <- as.vector(X%*%beta[,h])

}
}

#-------------------------



# E-step

eobj<-tau.estep.wire(dat,pro,mu,BC,n,m,g)
pro   <-eobj$pro
tau   <-eobj$tau
lk[i] <-eobj$loglik
sumtau<- colSums(tau)


#-------------------------------

M<-matM.wire(B,C,tau,g,m)

if(ncov != "AR")
obj1 <- eb.estep(tau,dat,mu,thetu,U,B,C,M,g,n,m,qb)
else
obj1 <- ar.wire.estep(tau,dat,mu,thetu,B,C,M,n,m,g)


obj2 <- ec.estep(tau,dat,mu,thetv,V,M,g,n,qc)
obj3 <- ee.estep(dat,mu,tau,U,V,W,A,obj1$invB,M,g,n,m,qe,dw,obj1$eb1,obj2$ec1)


# M-step

if(ncov != "AR")
{
if(ncov>0)
thetu<-obj1$DU
else
for(h in 1:g)
thetu[,,h]<-diag(0,qb)

if( (ncov>0) & (ncov!=3)&(ncov!="AR") )
thetu <- getcov(thetu,sumtau,n,qb,g,ncov)


}else {

if(!common) {
moo <- ar.wire.mstep(m,g,obj1$TU,sumtau)
theta <- moo$theta
phi   <- moo$phi
} else {

etyy <- array(0,c(m,m,1))

for(h in 1:g)
etyy[,,1] <- etyy[,,1] + obj1$TU[,,h]

moo   <- ar.wire.mstep(m,1,etyy,c(sum(sumtau)))
theta <- rep(moo$theta[1],g)
phi   <- rep(moo$phi[1],g)

}

}


if(nvcov>0)
thetv<-obj2$ec2
else
thetv<-rep(0,qc)

thete<-obj3$thete


#bestperiod <- get.period(dat,beta,g,A,tau,tim,k,one,periods,U,V,obj1$eb1,obj2$ec1)

#--------------------------------

for(h in 1:g)
beta[,h]<-(beta[,h]+xxx[,,h]%*%M[,,h]%*%A[,,h]%*%colSums(t(t(dat)-mu[,h])*tau[,h])/sumtau[h])


#

if(debug)
{
cat('\n',i,'th,loglik=',lk[i])
if(ncov=="AR")
cat('\ntheta=',theta,'\nrho=',phi)
if(tmtype)
cat('\nbestperiod=',bestperiod)

}

loglik <- lk[i]

if(i<=100) next

if(abs(lk[i]-lk[i-1])< epsilon*abs(lk[i-1]))
#if((lk[i]-lk[i-10])< epsilon*abs(lk[i-10]))
{flag<-1;break}


} # end of loop

if(flag==0) error <- 1

#get the the final partition

for(h in 1:g) {

if(tmtype)
X     <- get.X(tim,bestperiod[h],k,one)

mu[,h]<-as.vector(X%*%beta[,h]+V%*%obj2$ec1[,h])
}

eobj2<-tau.estep.wire(dat,pro,mu,B,n,m,g)
clust<-tau2clust(eobj2$tau)

#
# BIC & AIC
p<-qb;

nu<-switch(paste(ncov),
    '0' = 0,              # without u
    '1' = p*(1+p)/2,     #common covariance
    '2' = p,             #common diagonal covariance
    '3' = g*(p*(1+p)/2),#general covariance
    '4' = g*p,          #general diagonal covariance
    'AR'= g*2,          #AR(1) model
    '5' = g  )          #sigma(h)*I_p
nv<-ifelse(nvcov>0,g*qc,0)

np<-(g-1)+nb*g + nu + nv+ g*qe
#
BIC<--2*loglik+np*log(n)
AIC<--2*loglik+np*2
if(debug)
cat('\n',g,"BIC=",BIC,"AIC=",AIC,loglik,np)


#return values

ret <- list(error=error,loglik=loglik,BIC=BIC,AIC=AIC,pro=pro,beta=beta,mu=mu,eb1=obj1$eb1,ec1=obj2$ec1)



if(ncov==1|ncov==2|ncov==3|ncov==4|ncov==5)
{
ret$sigma  <- thete
ret$thetab <- thetu
if(nvcov >0)
ret$thetac <- thetv
}


if(ncov=="AR") {
ret$sigma2  <- thete
ret$theta2  <- theta
ret$rho     <- phi

if(nvcov >0)
ret$d2      <- thetv

ret$sigma  <- thete
ret$theta  <- theta
if(nvcov >0)
ret$thetac <- thetv


}


if(tmtype) {
ret$bestperiod <- bestperiod


}

ret$lk =lk[lk!=0]
ret$clust=clust
ret$tau=eobj2$tau

ret
}


# fit the g components models

# start from initial partition

emmixwire<-function(dat,g,ncov=3,nvcov=1,tmtype=FALSE,one=FALSE,common=TRUE,
X=NULL,W=NULL,U=NULL,V=NULL,clust=NULL,
debug=0,itmax=1000,epsilon=1e-3,nkmeans=5,
tmpoints=NULL,k=1,periods=NULL,initperiod=NULL,...)
{

dat<-as.matrix(dat)
n<-nrow(dat)
m<-ncol(dat)

# check the matrix U

if(ncov =="AR"|ncov==0)
{
U     <- diag(m)

} else {

if(ncov>0 & is.null(U))
U <- cbind(rep(1,m))

}

# check the matrix V

if(is.null(V) | nvcov==0)
V  <- diag(m)




# check the matrix X

if(tmtype)
X <- get.X(tmpoints,initperiod[1],k,one)
else {

if(is.null(X))
stop("X must be specified")

}


# check the matrix W

if(is.null(W))
W <- cbind(rep(1,m))


#
U<-as.matrix(U)
V<-as.matrix(V)
W<-as.matrix(W)
X<-as.matrix(X)


#  some variables

nb<-ncol(X)
qb<-ncol(U)
qc<-ncol(V)
qe<-ncol(W)

#

theta <- rep(1,g)
phi   <- rep(0.1,g)

tuv <- 0.2

# initialize the theta_b and theta_c

thetu<-array(0,c(qb,qb,g))

for(h in 1:g)
{
if(qb>1)
diag(thetu[,,h])<-rep(tuv,qb)
else
thetu[1,1,h]<- tuv
}

thetv<-rep(tuv,g)



#part 1: initial values

if(is.null(clust))
found <- ar.init.fit(dat,X,qe,n,m,g,nkmeans)
else
found <- ar.init.reg(dat,X,qe,n,m,g,clust)


#part 2: call the main estimate procedure

ret<-fit.emmix.wire(dat,X,W,U,V,
found$pro,found$beta,found$thete,thetu,thetv,theta,phi,
n,m,g,nb,qb,qc,qe,
debug,ncov,nvcov,itmax,epsilon,
tmtype,tmpoints,periods,k,one,initperiod,common,mu,eb1,ec1)


if(qb==m & (ncov==4 | ncov==2))
{
tmp <- array(0,c(m,g))

for(h in 1:g)
tmp[,h] <- diag(ret$thetab[,,h])

ret$thetab <- tmp

}


ret
}



#---------------------------------
# end of program
#---------------------------------




