

# Computation of pairwise Spearman's correlation
cal.measure<-function(data,nv,measure="spearman") {
n<-nrow(data)
nv<-ncol(data)-1
corr<-cor(data,use="complete.obs",method=measure)
corr
}


# permutation 
# Note: Example 8.1 used measure = "pearson" here incorrectly
permuteit<-function(data,n,nv,B,measure="spearman") {
rand<-array(0,c(n,nv))
corrnull<-array(0,c(nv,nv))
pairnull<-array(0,c(nv*(nv-1)/2*B))
set.seed(1)
randint<-sample(1:20,5)
if (B>100) {
  rr<-floor(B/20)+1
  randint<-c(randint,sample(21:B,rr))
}
count<-0
jj<-1
for (k in 1:B) {                           # number of repetition B   
  if ((k-1)%%20==0) {
    count<-count+1
    set.seed(randint[count])
  }
  for (i in 1:nv) rand[,i]<-data[sample(1:n,n,replace=F),i]   # permutation
  corrnull<-cal.measure(rand,nv,measure)      # null Somers D based on permuted data 
  for (row in 1:(nv-1)) {
    for (col in (row+1):nv) {
      pairnull[jj]<-corrnull[row,col]
      jj<-jj+1
    }
  }
}
pairnull
}


# determine cut-off value with control for FDR
cutit<-function(pairnull,corr,nv,alpha) {
len<-length(pairnull)
ni<-nv*(nv-1)/2
pairr<-array(0,c(ni,4))
pairrsort<-array(0,c(ni,4))
jj<-1
for (row in 1:(nv-1)) {
  for (col in (row+1):nv) {
    pairr[jj,1]<-abs(corr[row,col])
    pairr[jj,2]<-row
    pairr[jj,3]<-col
    pairr[jj,4]<-sum(abs(pairnull)>=abs(corr[row,col]))/(len)
    jj<-jj+1
  }
}
pairrsort<-pairr[c(rev(order(pairr[,1]))),]
cutoff<-0
for (jj in 1:ni) {
  if (pairrsort[jj,4]<=alpha*jj/ni & pairrsort[jj,1]>=0.1) {             
    cutoff<-jj                                      # FDR control at alpha
    cutcorr<-pairrsort[jj,1]
  }
}
list(cutoff=cutoff,cutcorr=cutcorr,pairrsort=pairrsort)
}


# Computation of the M-matrix
cal.matrix<-function(corr,nv,cutcorr) {
Mmatrix<-array(0,c(nv,nv))
for (row in 1:(nv-1)) {
  for (col in (row+1):nv) {
    if (abs(corr[row,col])>=abs(cutcorr)) {
      Mmatrix[row,col]<-sign(cutcorr)
      Mmatrix[col,row]<-Mmatrix[row,col]          # compute M matrix 
    }
  }
}
Mmatrix
}

# Clustering
clustit<-function(corr,pairrsort,Mmatrix,nv,cutoff,cutcorr) {
gp<-array(0,c(cutoff,nv))
strength<-rep(0,cutoff)
positive<-rep(1,cutoff)
negative<-rep(0,cutoff)
av<-rep(0,nv)
sump<-rep(0,nv)
sumn<-rep(0,nv)
clust<-0
for (cter in 1:cutoff) {
#  cat("iter: ", cter, "\n")
  ind<-1
  exist<-0
#  while (clust>=1 && ind<=clust) {                  # search the M matrix
#    if (sum(gp[ind,]==pairrsort[cter,2])==1 && 
#    sum(gp[ind,]==pairrsort[cter,3])==1) { exist=1;break }
#    ind<-ind+1   }
#  if (exist==1) next
  clust<-clust+1
  gp[clust,1]<-pairrsort[cter,2]
  gp[clust,2]<-pairrsort[cter,3]
  strength[clust]<-strength[clust]+abs(corr[gp[clust,1],gp[clust,2]])
#  if (corr[gp[clust,1],gp[clust,2]]>0.) {
#    positive[clust]<-positive[clust]+sign(corr[gp[clust,1],gp[clust,2]])
#  } else {
#    negative[clust]<-negative[clust]+sign(corr[gp[clust,1],gp[clust,2]])  }
  num<-2
  ref<-1.
  while (ref>0.) { 
    for (k in 1:nv) {
      av[k]<-0
      sump[k]<-0
      sumn[k]<-0
      same<-1
      for (i in 1:num) {
        if (Mmatrix[gp[clust,i],k]==0) { same<-0;break }    }
      if (same==1) {  
        for (i in 1:num) {
          av[k]<-av[k]+abs(corr[k,gp[clust,i]])/num  
          if (corr[k,gp[clust,i]]>0.) {sump[k]<-sump[k]+sign(corr[k,gp[clust,i]]) }
          else {sumn[k]<-sumn[k]+sign(corr[k,gp[clust,i]]) }  }   }
    }
    ref<-0.
    refp<-0
    refn<-0
    for (k in 1:nv) {
      if (av[k]>ref) {
        ref<-av[k]
        refp<-sump[k]
        refn<-sumn[k]
        gp[clust,num+1]<-k
        }     #  break   }
    }
    strength[clust]<-strength[clust]+ref*num
    positive[clust]<-positive[clust]+refp
    negative[clust]<-negative[clust]+refn
    num<-num+1
  }
  if (num<4 && cutoff>1) {
    strength[clust]<-0.
    clust<-clust-1
    next   }
  gp[clust,1:(num-1)]<-sort(gp[clust,1:(num-1)])
  strength[clust]<-strength[clust]*2/((num-1)*(num-2))
}
if (clust>1) {
  gp<-gp[1:clust,]
  ret<-cbind(gp,strength[1:clust],positive[1:clust],negative[1:clust])
  if (nv>=8) {
  ret<-ret[order(gp[,1],gp[,2],gp[,3],gp[,4],gp[,5],gp[,6],gp[,7],gp[,8]),]
  } else {   
  ret<-ret[order(gp[,1],gp[,2],gp[,3]),]  }
  ret<-cbind(ClusterID=1:clust,ret) 
  } else { ret<-matrix(c(ClusterID=1,gp[1,],strength[1]),ncol=nv+2) }
dimnames(ret)<-list(1:clust,c("ClusterID",rep(" ",nv),"Strength","Positive","Negative"))
ret  
}

# Extraction of co-morbidity clusters
extract<-function(ret,nv,ofinterest) {
clust<-nrow(ret)
bad<-0
if (clust>1) {
  for (j in 1:clust) {
    if (sum(ret[j,2:(nv+1)]==ofinterest)==1) next
    bad<-c(bad,j)  
    }
  len<-length(bad)
  if (len>1) {
    bad<-bad[-1]
    ret<-matrix(ret[(1:clust)[-bad],],ncol=nv+2)
    clust<-clust-len+1
    ret[,1]<-1:clust  
    dimnames(ret)<-list(1:clust,c("ClusterID",rep(" ",nv),"Strength"))
  }
}
ret
}

